// -----------------------------------------------------------------------
// Litmus - DKFZ SIDT unit testing library
//
// Copyright (c) German Cancer Research Center (DKFZ),
// Software development for Integrated Diagnostic and Therapy (SIDT).
// ALL RIGHTS RESERVED.
// See litmusCopyright.txt or
// http://www.dkfz.de/en/sidt/projects/litmus/copyright.html
//
// This software is distributed WITHOUT ANY WARRANTY; without even
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.  See the above copyright notices for more information.
//
//------------------------------------------------------------------------
/*!
// @file
// @version $Revision: 4433 $ (last changed revision)
// @date    $Date: 2013-01-22 14:38:20 +0100 (Di, 22 Jan 2013) $ (last change date)
// @author  $Author: floca $ (last changed by)
*/


#ifndef __LIT_FIELD_TESTER_TPP
#define __LIT_FIELD_TESTER_TPP

#include "itkImageRegionConstIterator.h"
#include <assert.h>

namespace lit 
{
    template<class TField>
    StringType
      FieldTester<TField>::
      getTestDescription(void) const
    {
      return "Checks if actual vector field equals another field";
    };

    template<class TField>
    StringType
      FieldTester<TField>::
      getTestName(void) const
    {
      OStringStream stream;
      stream << "Check equal: actual vector field == expected vector field (tolerance: +-"<<_checkThreshold<<")";

      return stream.str();
    };

    template<class TField>
    void
      FieldTester<TField>::
      setExpectedField(const FieldType* pField)
    {
      _spExpectedField = pField;
    };

    template<class TField>
    void
      FieldTester<TField>::
      setActualField(const FieldType* pField)
    {
      _spActualField = pField;
    };

    template<class TField>
    void
      FieldTester<TField>::
      setCheckThreshold(double checkThreshold)
    {
      _checkThreshold = checkThreshold;
    };

    template<class TField>
    const typename FieldTester<TField>::FieldType*
      FieldTester<TField>::
      getExpectedField() const
    {
      return _spExpectedField;
    };

    template<class TField>
    const typename FieldTester<TField>::FieldType*
      FieldTester<TField>::
      getActualField() const
    {
      return _spActualField;
    };


    template<class TField>
    double
      FieldTester<TField>::
      getCheckThreshold() const
    {
      return _checkThreshold;
    };

    template<class TField>
    bool
    FieldTester<TField>::
      getCheckFieldGeometry() const
    {
      return _checkGeometry;
    };

    template<class TField>
    void
    FieldTester<TField>::
      setCheckFieldGeometry(bool checkThreshold)
    {
      _checkGeometry = checkThreshold;
    };


    template<class TField>
    FieldTester<TField>::
      FieldTester() : _checkThreshold(0.0), _checkGeometry(true)
    {
    };

    template<class TField>
    FieldTester<TField>::
      ~FieldTester()
    {
    };

    /*! performes the test and checks the results.
    * @result Indicates if the test was successfull (true) or if it failed (false)
    */
    template<class TField>
    bool
      FieldTester<TField>::
      doCheck(void) const
    {
      bool fieldEqualsField = true;
      _maxError = 0.0;
      _errorPixelCount = 0;
      _maxErrorIndex.Fill(0);

      _pResults->onTestStart(getCurrentTestLabel());

      if (_checkGeometry)
      {
        if (!(_spExpectedField->GetSpacing() == _spActualField->GetSpacing()))
        {
          fieldEqualsField = false;
        }
        if (!(_spExpectedField->GetOrigin() == _spActualField->GetOrigin()))
        {
          fieldEqualsField = false;
        }
        if (!(_spExpectedField->GetDirection() == _spActualField->GetDirection()))
        {
          fieldEqualsField = false;
        }
      }

      if (_spExpectedField->GetLargestPossibleRegion() == _spActualField->GetLargestPossibleRegion())
      {
      //check for every field pixel if it maps like the transform model.
      typedef itk::ImageRegionConstIterator< FieldType > IteratorType;
      IteratorType expectedIterator(_spExpectedField,_spExpectedField->GetLargestPossibleRegion());
      IteratorType actualIterator(_spActualField,_spActualField->GetLargestPossibleRegion());
      actualIterator.GoToBegin();
      for (expectedIterator.GoToBegin(); !(expectedIterator.IsAtEnd()); ++expectedIterator, ++actualIterator)
      {
        typename FieldType::PixelType differenceVector = expectedIterator.Get() - actualIterator.Get();
        double difference = differenceVector.GetNorm();

        if (difference>_checkThreshold)
        {
          fieldEqualsField = false;
          ++_errorPixelCount;
          if (_maxError<difference)
          {
            _maxError = difference;
            _maxErrorIndex = expectedIterator.GetIndex();
          };
        }
      }

      }
      else
      {
        fieldEqualsField = false;
      }

      return fieldEqualsField;
    };

    template<class TField>
    void
      FieldTester<TField>::
      handleSuccess(void) const
    {
      OStringStream stream;
      stream << "Vectors of "<<_spActualField << std::endl <<"equals (tolerance: "<<_checkThreshold<<")"<<std::endl<<std::endl<< _spExpectedField;
      _pResults->onTestSuccess(getCurrentTestLabel(), stream.str());
    };

    template<class TField>
    void
      FieldTester<TField>::
      handleFailure(void) const
    {
      OStringStream stream;
      stream << "Vectors of "<<_spActualField << std::endl << " doesn't equal (tolerance: "<<_checkThreshold<<") "<<std::endl<<std::endl<< _spExpectedField << std::endl;
      if (!(_spExpectedField->GetLargestPossibleRegion() == _spActualField->GetLargestPossibleRegion()))
      {
        stream << std::endl << "Largest possible regions of vector fields are not equal." << std::endl;
      }
      stream << std::endl << "Error pixel count:   "<< _errorPixelCount << std::endl;
      stream << std::endl << "Maximum difference:  "<< _maxError << std::endl;
      stream << std::endl << "Error pixel index:   "<< _maxErrorIndex << std::endl;
      stream << std::endl << "<DartMeasurement name=\"Error pixel count\" type=\"numeric/int\">"<< _errorPixelCount << "</DartMeasurement>";
      stream << std::endl << "<DartMeasurement name=\"Maximium error (difference)\" type=\"numeric/double\">"<< _maxError << "</DartMeasurement>";
      stream << std::endl << "<DartMeasurement name=\"Error pixel index\" type=\"string\">"<< _maxErrorIndex << "</DartMeasurement>" << std::endl;

      _pResults->onTestFailure(getCurrentTestLabel(), stream.str());
    };

}

#endif
